options(stringsAsFactors=FALSE)
library(matlab)
library(parallel)
library(pls)
library(glmnet)
library(rpart)
#library(RhpcBLASctl) # possibly uncomment this line
library(tidyverse)
library(scales)
library(ggdendro)
library(cowplot)

rm(".Random.seed")

load("parse_raw_data.RData")


source(file="predict_models.R")
source(file="settings.R")

################################
######## HELPER FUNCTIONS ######
################################

R2 = function(meas,preds) {
  notNA = !is.na(meas) & !is.na(preds)
  meas = meas[notNA]
  preds = preds[notNA]
  return(1-sum((meas-preds)^2)/sum((meas-mean(meas))^2))
}

list2str = function(l) {
  if('tissue' %in% names(l))
    if(length(l$tissue)>1)
      l$tissue = paste(l$tissue, collapse="-")
    namesort = c("datasource", "datatype", "tissue", "Tx", "version")
    namesort = ordered(namesort, levels=namesort)
    l = l[  paste(sort(factor(names(l), levels=namesort, ordered = T))) ]
    return(paste(names(l), toupper(l), collapse="_", sep=""))
}

list2str_reduce = function(x,y) {
  return(paste(list2str(x),list2str(y),sep="_"))
}

listlist2str = function(l) {
  if(length(l)==0)
    return("")
  else if(length(l)>0 & typeof(l[[1]])!="list")
    return("INCORRECT_DATA_FORMAT")
  else {
    return(paste(lapply(l,list2str), collapse="_"))
  }
}

convert2ListOfList = function(l) {
  if(length(l)>0)
    if(typeof(l[[1]])!="list")
      return(list(l))
  return(l)
}

str_wrap_factor <- function(x, ...) {
  levels(x) <- str_wrap(levels(x), ...)
  return(x)
}


################################
##### PREDICT DRUG RESPONSE ####
################################
predict_wrapper = function(dataX, 
                      dataY, 
                      dataY_subset=NULL, 
                      method = "pls", 
                      match_CLs=NULL, 
                      Nrep, 
                      Nbg = 0, 
                      Ncore = 1, 
                      NcoreBLAS = 1,
                      remove_dups=F, 
                      exclude_CLs=NULL, 
                      integration = 'early', 
                      filt_settings = NULL,
                      return_model=F, 
                      return_preds=F,
                      blas_cores_limit=T) {
  
  
  
  ###############################################
  ##### LOAD SETTINGS AND DATA SOURCE ###########
  ###############################################
  
  # set number of threads to use for linear algebra package
  # possibly uncomment below 4 lines
  #if(blas_cores_limit) {
  #  omp_set_num_threads(1)
  #  blas_set_num_threads(1)
  #}
  
  # constants
  MUT_NMIN = 3
  MIN_NOBS = 17
  MIN_DRUG_SD = 0.2
  
  # make dataX/filt/match_CLs a list of inputs
  dataX = convert2ListOfList(dataX)
  filt_settings = convert2ListOfList(filt_settings)
  match_CLs = convert2ListOfList(match_CLs)
  match_CLs_ext = match_CLs 
  

  ###############################################
  ########## BUILD UP X AND Y DATA  #############
  ############################################### 
  data_xs = list()
  match_CLs = c(match_CLs, dataX); match_CLs[[length(match_CLs)+1]]=dataY
  for(dxi in seq(1,length(dataX))) {
    dx = dataX[[dxi]]
    df = getData(dx, remove_dups=remove_dups, match_CLs=match_CLs)
    if(nrow(df)<MIN_NOBS | is.null(df)) {
      print("Too little data in matched x-data, quitting.")
      return(NULL)
    }
    if(length(filt_settings)==length(dataX) & !listItemEquals(dx, 'version', 'unfilt')) {
      sd_filt = filt_settings[[dxi]]$sd_filt
      mean_filt = filt_settings[[dxi]]$mean_filt
    } else {
      sd_filt = 0.1 # there has to be at least a little variation
      mean_filt = -999
      if(listItemEquals(dx, 'version', 'unfilt')) {
        filt_settings[[dxi]]$mean_filt = mean_filt
        filt_settings[[dxi]]$sd_filt = sd_filt
      }
    }
    # don't filter input data if it has character types, or if cosmic/random genes were selected
    if(!any( sapply(df,typeof)=="character") & dx$datatype!="mut" & !listItemContains(dx,'version', 'random') & !listItemContains(dx,'version', 'cosmic'))
      df = df[,apply(df,2,function(x) sd(x, na.rm=T))>sd_filt & apply(df,2,function(x) mean(x, na.rm=T))>mean_filt, drop=F]
    else if(dx$datatype=="mut") {
      nmuts = apply(df,2,function(x) sum(x))
      df = df[, (nmuts>=MUT_NMIN) & (nmuts<=nrow(df)-MUT_NMIN) ]
    }
    if(!is.null(exclude_CLs))
      df = df[!(rownames(df) %in% exclude_CLs),,drop=F]
    data_xs[[length(data_xs)+1]] = df
  }
  
  if(length(data_xs)==0) { 
    print("ERROR: No x-data.")
    return(NULL)
  }
  if(length(data_xs)>2 & integration=="late") { 
    print("ERROR: Late integration only supported for 2 input dataset.") 
    return(NULL)
  }
  if(length(data_xs)==1) {
    data_x = data_xs[[1]]
  } else if(length(data_xs)==2 & integration=="late") {
    data_x1 = data_xs[[1]]
    data_x2 = data_xs[[2]]
  } else if(integration=="early") {
    data_x=reduce(data_xs, cbind) # 
  }
  
  data_y = getData(dataY, remove_dups=remove_dups, match_CLs=match_CLs)
  if(!is.null(dataY_subset)) {
    data_y = data_y[,colnames(data_y) %in% dataY_subset, drop=F]
  }
  if(!is.null(exclude_CLs))
    data_y = data_y[!(rownames(data_y) %in% exclude_CLs),,drop=F]
  
  # remove drugs with too many missing values
  keep = apply(data_y,2, function(x) length(which(!is.na(x)))>=MIN_NOBS)
  data_y = data_y[, keep , drop=F]
  
  # quit if too few cell lines
  if(nrow(data_y)<MIN_NOBS | ncol(data_y)==0) {
    print("Too little data, quitting.")
    return(NULL)
  }
  
  # Remove Y variables with too many missing values
  # or too little variance. Don't do for gt data,
  # of if drugs are explicitly provided.
  if(dataY$datasource != "gt" & is.null(dataY_subset)) {
    nObs = apply(data_y,2,function(x) length(which(!is.na(x))))
    sds = apply(data_y,2,function(x) sd(x,na.rm=T)/mean(x,na.rm=T))
    data_y = data_y[,nObs>MIN_NOBS & sds>MIN_DRUG_SD, drop=FALSE]
    if(ncol(data_y)==0) {
      print("Too little Y-data")
      return(NULL)
    }
  }
  
  if(integration=="early") {
    print(paste("X-dim: ", toString(nrow(data_x)), "x", toString(ncol(data_x)),", Y-dim:", 
                toString(nrow(data_y)), "x", toString(ncol(data_y)), sep=""))
  }
  
  # Check that cell lines for x/y data match up
  for(i in seq(1,length(data_xs))) {
    if(nrow(data_xs[[i]]) != nrow(data_y) |  any(rownames(data_xs[[i]]) != rownames(data_y))) {
      print("ERROR: CELL LINE NAMES MUST MATCH UP FOR X AND Y DATA")
      return(NULL)
    }
  }
  
  
  ##############################################
  ########## PREDICT VIABILITY #################
  ##############################################
  
  cluster <- makeCluster(min(Nrep,Ncore), type="FORK")
  clusterEvalQ(cluster,library(pls))
  clusterEvalQ(cluster,library(glmnet))
  clusterEvalQ(cluster,library(rpart))
  
  # repeat predicitons (average over many K-fold CVs)
  seed_offset = round(runif(1,0,10^6))
  outN = parLapply(cluster,seq(1,Nrep),function(x) { 
    
    if(method=="pls") {
      if(integration == 'late' & exists("data_x1") & exists("data_x2")) {
        out = runpls_twosources(data_x1,data_x2,data_y, shuffle=FALSE, seed=seed_offset+x)
      } else {
        out = runpls(data_x,data_y, shuffle=FALSE, seed=seed_offset+x, return_model = return_model, return_preds = return_preds)
      }
    } else if(method=="plsmulti") {
      out = runplsmulti(data_x,data_y, shuffle=FALSE, seed=seed_offset+x)
    } else if(method=="maxcor") {
      out = runmaxcor(data_x,data_y, shuffle=FALSE, seed=seed_offset+x)
    } else if(method=="maxcor") {
      out = runmaxcor(data_x,data_y, shuffle=FALSE, seed=seed_offset+x)
    } else if(method=="dummy") {
      out = rundummy(data_x,data_y, shuffle=FALSE, seed=seed_offset+x)
    } else if(method=="elasticnet") {
      out = runelasticnet(data_x,data_y,shuffle=FALSE, seed=seed_offset+x)
    } else if(method=="lasso") {
      out = runelasticnet(data_x,data_y, alphas=c(1.0), shuffle=FALSE, seed=seed_offset+x)
    } else if(method=="tissue") {
      out = runtissue(data_x,data_y, shuffle=FALSE, seed=seed_offset+x)
    } else if(method=="regressiontree") {
      out = runregressiontree(data_x,data_y, shuffle=FALSE, seed=seed_offset+x)
    }
    return(out)
    
    } )
  stopCluster(cluster)
  
  # save results to output variable
  out = list()
  if(is.null(dim(sapply(outN,function(x) x$varsExpl)))) {  # if only one drug
    out$varsExpl_rep = sapply(outN,function(x) x$varsExpl)
    out$cors_rep = sapply(outN,function(x) x$cors)
    varsExpl = mean(out$varsExpl_rep)
    cors = mean(out$cors_rep)
  } else {  # multiple drugs
    out$varsExpl_rep = t(sapply(outN,function(x) x$varsExpl))
    out$cors_rep = t(sapply(outN,function(x) x$cors))
    varsExpl = apply(out$varsExpl_rep,2,mean)
    cors = apply(out$cors_rep,2,mean)
  }
  out$cors = cors
  out$varsExpl = varsExpl
  if('data_y' %in% names(outN[[1]])) {
    out$data_y = outN[[1]][['data_y']]
    out$preds_array = sapply(outN, function(x) x$preds, simplify="array")
    out$preds = apply(out$preds_array, 1:2, mean)
  }
  if('models' %in% names(outN[[1]])) {
    out$models = outN[[1]]$models
  }
  
  
  
  ###############################################
  ###### VIABILITY BACKGROUND MODEL #############
  ###############################################
  if(Nbg>0) {
    print("=== COMPUTING BACKGROUND DISTRIBUTIONS ===")
    
    seed_offset = round(runif(1,0,10^6)) 
    cluster <- makeCluster(Ncore, type="FORK")
    clusterEvalQ(cluster,library(pls))
    clusterEvalQ(cluster,library(glmnet))
    clusterEvalQ(cluster,library(rpart))
    if(method=="pls" && integration=="early") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { runpls(data_x,data_y,shuffle=TRUE,seed=x+seed_offset, return_model=F) } )
    } else if(method=="pls" && integration=="late") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { runpls_twosources(data_x1,data_x2,data_y, shuffle=TRUE, seed=x+seed_offset) } )
    } else if(method=="plsmulti") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { runplsmulti(data_x, data_y, shuffle=TRUE, seed=x+seed_offset) } )
    } else if(method=="maxcor") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { runmaxcor(data_x,data_y,shuffle=TRUE,seed=x+seed_offset) } )
    } else if(method=="elasticnet") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { runelasticnet(data_x,data_y,shuffle=TRUE,seed=x+seed_offset) } )
    } else if(method=="lasso") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { runelasticnet(data_x,data_y,alphas=c(1.0), shuffle=TRUE,seed=x+seed_offset) } )
    }  else if(method=="dummy") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { rundummy(data_x,data_y,shuffle=TRUE,seed=x+seed_offset) } )
    } else if(method=="tissue") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { runtissue(data_x,data_y,shuffle=TRUE,seed=x+seed_offset) } )
    } else if(method=="regressiontree") {
      bgdata = parLapply(cluster,seq(1,Nbg),function(x) { runregressiontree(data_x,data_y,shuffle=TRUE,seed=x+seed_offset) } )
    }
    stopCluster(cluster)
    
    cors_bg = t(sapply(bgdata,function(x) x$cors))
    varsExpl_bg = t(sapply(bgdata,function(x) x$varsExpl))
    out$cors_bg = cors_bg
    out$varsExpl_bg = varsExpl_bg
    
    x_names = colnames(data_y)
    
    # compute p-values
    pvals_cors = c()
    pvals_varsExpl = c()
    for(d in seq(1,length(x_names))) {
      pval_cor = min(1, (length(which(cors_bg[,d]>cors[d]))+1)/length(cors_bg[,d]) )
      pval_varExpl = min(1, (length(which(varsExpl_bg[,d]>varsExpl[d]))+1)/length(varsExpl_bg[,d]) )
      pvals_cors = c(pvals_cors,pval_cor)
      pvals_varsExpl = c(pvals_varsExpl,pval_varExpl)
    }
    names(pvals_cors) = x_names
    names(pvals_varsExpl) = x_names
    pvals_cors_adj = p.adjust(pvals_cors,method="BH")
    pvals_varsExpl_adj = p.adjust(pvals_cors,method="BH")
    
    out$pvals_cors = pvals_cors
    out$pvals_cors_adj = pvals_cors_adj
    out$pvals_varsExpl = pvals_varsExpl
    out$pvals_varsExpl_adj = pvals_varsExpl_adj
    
  } 
  
  out$filt_settings = filt_settings
  dt = getDrugTarget(dataY, remove_dups=remove_dups, match_CLs=match_CLs)
  m = match(names(out$cors), dt$drugs)
  if(any(is.na(m))) {
    print("Missing drug type information, quitting.")
    return(NULL)
  }
  m = m[!is.na(m)]
  if(!is.null(dt$drug_targets)) {
    out$drug_targets = dt$drug_targets[m]
    names(out$drug_targets) = names(out$cors)
  }
  out$dataX = dataX
  out$dataY = dataY
  out$match_CLs_ext = match_CLs_ext
  if(length(dataX)==1 & is.null(match_CLs_ext)) {
    out$argstr = paste("predict_wrapper_",listlist2str(dataX),"_",list2str(dataY), "_", method,"_meanFilt",mean_filt,"_sdFilt",sd_filt,"_removedups",toString(remove_dups),"_Nbg",Nbg,sep="")
  } else if(length(dataX)>=1 & is.null(match_CLs_ext)) {
    out$argstr = paste("predict_wrapper_",listlist2str(dataX),"_",list2str(dataY), "_", method,"_meanFilt",mean_filt,"_sdFilt",sd_filt,"_integration",toupper(integration),"_removedups",toString(remove_dups),"_Nbg",Nbg,sep="")
  } else if(length(dataX)==1 & !is.null(match_CLs_ext)) {
    out$argstr = paste("predict_wrapper_",listlist2str(dataX),"_",list2str(dataY), "_matched_", method,"_meanFilt",mean_filt,"_sdFilt",sd_filt,"_removedups",toString(remove_dups),"_Nbg",Nbg,sep="")
  } else if(length(dataX)>=1 & !is.null(match_CLs_ext)) {
    out$argstr = paste("predict_wrapper_",listlist2str(dataX),"_",list2str(dataY), "_matched_", method,"_meanFilt",mean_filt,"_sdFilt",sd_filt,"_integration",toupper(integration),"_removedups",toString(remove_dups),"_Nbg",Nbg,sep="")
  }
  
  return(out)
  
} # main function

